using System;
using System.IO;			// for StreamWriter I/O to a file
using System.Windows.Forms;	// for RichTextBox

namespace MyNamespace
{
	/// <summary>
	///		Multiple precision signed integer Matrix algorithms
	///		in Visual C# .Net
	///
	/// ---------------- *** LEGAL NOTICE: *** ------------------
	///
	/// This program is free software; you can redistribute it and/or modify
	/// it under the terms of the GNU General Public License (GPL) as published
	/// by the Free Software Foundation. cf. the file COPYING.txt.
	///
	/// ------------ *** end of legal notice *** ------------------
	/// </summary>

	// ------- A multiple-precision signed integer matrix
	public class MultiJ : MultiSI	// array of signed integers
	{
		public int Rs;		// Number of rows
		public int Cs;		// Number of columns
		public MultiSI[,] E;	// elements

		public MultiJ()				// initialize a scalar with M = 1
			: base()				// calls MultiSI()
		{
			// constructor logic
			this.Rs = 0;
			this.Cs = 0;
			this.E = new MultiSI[0, 0];
		}

		public MultiJ(int nMaxI)		// initialize a scalar with M = nMaxI
			: base(nMaxI)				// calls MultiSI(i)
		{
			// constructor logic
			this.Rs = 0;
			this.Cs = 0;
			this.E = new MultiSI[0, 0];
		}

		public MultiJ(int nMaxI, int rs, int cs)	// initialize a matrix
			: base(0)								// calls MultiSI(0)
		{
			// constructor logic
			if (rs < 1 || cs < 1)
			{
				rs = cs = 0;						// change to a scalar
				this.V = new int[nMaxI+1];
				this.M = nMaxI;
				this.N = 1;
			}
			this.Rs = rs;
			this.Cs = cs;
			this.E = new MultiSI[rs, cs];
			for (int r = 0; r < rs; r++)
				for (int c = 0; c < cs; c++)
				{
					this.E[r, c] = new MultiSI(nMaxI);
				}
		}

		//		public static bool operator = (MultiJ left, MultiJ right)	// not allowed to overload =
//		{
//			MultiJD.SetToM(right, left);
//			return true;
//		}

		public static MultiJ operator & (MultiJ left, MultiJ right)	// use &= for = assignment
		{
			MultiJD.SetToM(right, left);
			return left;
		}

		public static MultiJ operator & (MultiJ left, MultiSI right)	// use &= for = assignment
		{
			MultiJD.SetToSIM(right, left);
			return left;
		}

		public static MultiJ operator & (MultiJ left, int right)	// use &= for = assignment
		{
			MultiJD.SetTo1J(right, left);
			return left;
		}

		public static MultiJ operator & (MultiJ left, long right)	// use &= for = assignment
		{
			MultiJD.SetToExM(right, left);
			return left;
		}

		public static MultiJ operator & (MultiJ left, double right)	// use &= for = assignment
		{
			MultiJD.SetToDbM(right, left);
			return left;
		}

		public static MultiJ operator & (MultiJ left, decimal right)	// use &= for = assignment
		{
			MultiJD.SetToDeM(right, left);
			return left;
		}

		public static MultiJ operator + (MultiJ left, MultiJ right) // can be used for a += b
		{
			MultiJ z = new MultiJ(Calc.fMC, left.Rs, left.Cs);
			MultiJD.AddM(left, right, z);
			return z;
		}

		public static MultiJ operator + (MultiJ left, int right) // can be used for a += b
		{
			MultiJ z = new MultiJ(Calc.fMC);
			MultiJD.Add1M(left, right, z);
			return z;
		}

		public static MultiJ operator + (int left, MultiJ right) // can be used for a += b
		{
			MultiJ z = new MultiJ(Calc.fMC);
			MultiJD.Add1M(right, left, z);
			return z;
		}

		public static MultiJ operator - (MultiJ left, MultiJ right) // can be used for a -= b
		{
			MultiJ z = new MultiJ(Calc.fMC, left.Rs, left.Cs);
			MultiJD.SubM(left, right, z);
			return z;
		}

		public static MultiJ operator - (MultiJ left, int right) // can be used for a -= b
		{
			MultiJ z = new MultiJ(Calc.fMC);
			MultiJD.Add1M(left, -right, z);
			return z;
		}

		public static MultiJ operator - (int left, MultiJ right) // can be used for a -= b
		{
			MultiJ z = new MultiJ(Calc.fMC);
			MultiJD.Add1M(right, -left, z);
			MultiJD.ChSM(z);
			return z;
		}

		public static MultiJ operator - (MultiJ x)
		{
			MultiJ z = new MultiJ(x.N, x.Rs, x.Cs);
			MultiJD.SetToMS(x, z);
			MultiJD.ChSM(z);
			return z;
		}
		public static MultiJ operator * (MultiJ left, MultiJ right) // can be used for a *= b
		{
			MultiJ z = new MultiJ(Calc.fMC, left.Rs, right.Cs);
			MultiJD.MulM(left, right, z);
			return z;
		}

		public static MultiJ operator * (MultiJ left, int right) // can be used for a *= b
		{
			MultiJ z = new MultiJ(Calc.fMC, left.Rs, left.Cs);
			MultiJD.Mul1M(left, right, z);
			return z;
		}

		public static MultiJ operator * (int left, MultiJ right)
		{
			MultiJ z = new MultiJ(Calc.fMC, right.Rs, right.Cs);
			MultiJD.Mul1M(right, left, z);
			return z;
		}

		public static MultiJ operator / (MultiJ left, MultiJ right) // can be used for a /= b
		{
			int maxRs = Math.Max(left.Rs, right.Rs);
			int maxCs = Math.Max(left.Cs, right.Cs);
			MultiJ q = new MultiJ(Calc.fMC, maxRs, maxCs);
			MultiJD.DivM(left, right, q);
			return q;
		}

		public static MultiJ operator / (MultiJ left, int right) // can be used for a /= b
		{
			MultiJ q = new MultiJ(Calc.fMC, left.Rs, left.Cs);
			MultiJD.Div1M(left, right, q);
			return q;
		}

		// % and ^ will default to Singed Integer versions

		public static bool operator == (int d1, MultiJ z)
		{
			if (z.Rs != 0)
			{
				MultiJD.ErrIncompatible();
				return false;
			}
			MultiJ dF = new MultiJ(MultiID.sdInInt);
			MultiJD.SetTo1J(d1, dF);
			return (MultiID.CompSI(dF, z) == 0);
		}

		public static bool operator != (int d1, MultiJ z)
		{
			if (z.Rs != 0)
			{
				MultiJD.ErrIncompatible();
				return true;
			}
			MultiJ dF = new MultiJ(MultiID.sdInInt);
			MultiJD.SetTo1J(d1, dF);
			return (MultiID.CompSI(dF, z) != 0);
		}

		public static bool operator == (MultiJ z, int d1)
		{
			if (z.Rs != 0)
			{
				MultiJD.ErrIncompatible();
				return false;
			}
			MultiJ dF = new MultiJ(MultiID.sdInInt);
			MultiJD.SetTo1J(d1, dF);
			return (MultiID.CompSI(dF, z) == 0);
		}

		public static bool operator != (MultiJ z, int d1)
		{
			if (z.Rs != 0)
			{
				MultiJD.ErrIncompatible();
				return true;
			}
			MultiJ dF = new MultiJ(MultiID.sdInInt);
			MultiJD.SetTo1J(d1, dF);
			return (MultiID.CompSI(dF, z) != 0);
		}

		public static bool operator == (long ex, MultiJ z)
		{
			if (z.Rs != 0)
			{
				MultiJD.ErrIncompatible();
				return false;
			}
			MultiJ exF = new MultiJ(MultiID.sdInLong);
			MultiJD.SetToExM(ex, exF);
			return (MultiID.CompSI(exF, z) == 0);
		}

		public static bool operator != (long ex, MultiJ z)
		{
			if (z.Rs != 0)
			{
				MultiJD.ErrIncompatible();
				return true;
			}
			MultiJ exF = new MultiJ(MultiID.sdInLong);
			MultiJD.SetToExM(ex, exF);
			return (MultiID.CompSI(exF, z) != 0);
		}

		public static bool operator == (MultiJ z, long ex)
		{
			if (z.Rs != 0)
			{
				MultiJD.ErrIncompatible();
				return false;
			}
			MultiJ exF = new MultiJ(MultiID.sdInLong);
			MultiJD.SetToExM(ex, exF);
			return (MultiID.CompSI(exF, z) == 0);
		}

		public static bool operator != (MultiJ z, long ex)
		{
			if (z.Rs != 0)
			{
				MultiJD.ErrIncompatible();
				return true;
			}
			MultiJ exF = new MultiJ(MultiID.sdInLong);
			MultiJD.SetToExM(ex, exF);
			return (MultiID.CompSI(exF, z) != 0);
		}

		public static bool operator == (MultiJ x, MultiJ y)
		{
			return MultiJD.EqualM(x, y);
		}

		public static bool operator != (MultiJ x, MultiJ y)
		{
			return !MultiJD.EqualM(x, y);
		}

		public static bool operator < (MultiJ x, MultiJ y)
		{
			if (x.Rs != 0 || y.Rs != 0)
			{
				MultiJD.ErrNotScalar();
				return false;
			}
			return (MultiID.CompSI(y, x) < 0);
		}

		public static bool operator > (MultiJ x, MultiJ y)
		{
			if (x.Rs != 0 || y.Rs != 0)
			{
				MultiJD.ErrNotScalar();
				return false;
			}
			return (MultiID.CompSI(y, x) > 0);
		}

		public static bool operator <= (MultiJ x, MultiJ y)
		{
			if (x.Rs != 0 || y.Rs != 0)
			{
				MultiJD.ErrNotScalar();
				return false;
			}
			return (MultiID.CompSI(y, x) <= 0);
		}

		public static bool operator >= (MultiJ x, MultiJ y)
		{
			if (x.Rs != 0 || y.Rs != 0)
			{
				MultiJD.ErrNotScalar();
				return false;
			}
			return (MultiID.CompSI(y, x) >= 0);
		}

		public static MultiJ operator ++ (MultiJ x)
		{
			if (x.Rs != 0)
			{
				MultiJD.ErrNotScalar();
				return MultiJD.mZero;
			}
			MultiID.RAddSI(MultiID.sOne, x);
			return x;
		}

		public static MultiJ operator -- (MultiJ x)
		{
			if (x.Rs != 0)
			{
				MultiJD.ErrNotScalar();
				return MultiJD.mZero;
			}
			MultiID.RAddSI(MultiID.sMOne, x);
			return x;
		}

		public override bool Equals(object o)	// dummy
		{
			return true;
		}

		public override int GetHashCode()		// dummy
		{
			return 0;
		}
	}

	public class MultiJD
	{
		public MultiJD()	// constructor logic
		// initialize Multiple-precision integer matrix package
		// input: MultiID.muDMax, MultiID.muNMax, MultiID.baseM
		{
			mZero = new MultiJ(1);			// mZero = 0
			mOne = new MultiJ(1);
			SetTo1J(1, mOne);				// mOne = 1
			mTwo = new MultiJ(1);
			SetTo1J(2, mTwo);				// mTwo = 2
			mTen = new MultiJ(1);
			SetTo1J(10, mTen);				// mTwo = 10
			vZero = new MultiJ(1, 1, 1);	// vZero = {0}
		}

		public static bool EqualM(MultiJ x, MultiJ y)
		{
			if (x.Rs == 0 && y.Rs == 0)
				return (MultiID.CompSI(y, x) == 0);
			if (x.Rs != y.Rs || x.Cs != y.Cs)
			{
				ErrIncompatible();
				return false;
			}
			bool eq;
			for (int r = 0; r < x.Rs; r++)
				for (int c = 0; c < x.Cs; c++)
				{
					eq = (MultiID.CompSI(x.E[r, c], y.E[r, c]) == 0);
					if (!eq)
						return false;
				}
			return true;
		} // EqualM

		public static void Matrix(int nMaxI, int rs, int cs, out MultiJ z) // init matrix
		{
			z = new MultiJ(nMaxI, rs, cs);
		}

		public static void DoneJ(ref MultiJ z) // free up space
		{
			z = new MultiJ();
		}

		public static void ReSizeM(int m, ref MultiJ z) // change M and set to zero
		{
			z = new MultiJ(m, z.Rs, z.Cs);
		}

		public static void ClearM(ref MultiJ z)	// set Value = 0
		{
			z = new MultiJ(z.M, z.Rs, z.Cs);
		}

		public static void ChSM(MultiJ z) // change sign, z = - z
		{
			MultiID.ChSSI(z);
			for (int r = 0; r < z.Rs; r++)
				for (int c = 0; c < z.Cs; c++)
					MultiID.ChSSI(z.E[r, c]);
		}

		public static void SetToM(MultiJ x, MultiJ z)	// z = x
		// can reduce precision and round
		{
			int rs = x.Rs;
			int cs = x.Cs;
			if (rs != z.Rs || cs != z.Cs)
				ErrIncompatible();
			else
			{
				MultiID.SetToSI(x, z);
				for (int r = 0; r < rs; r++)
				{
					for (int c = 0; c < cs; c++)
					{
						MultiID.SetToSI(x.E[r, c], z.E[r, c]);
					}
				}
			}
		}

		public static void SetToSIM(MultiSI x, MultiJ z) // z = x
		// can reduce precision and round
		{
			if (z.Rs > 0)
			{
				z.E = new MultiSI[0, 0];
				z.Rs = z.Cs = 0;
			}
			MultiID.SetToSI(x, z);
		}

		public static void SetToMS(MultiJ x, MultiJ z) // z = x, same size
		// copy everything but .Nm Name
		{
			int rs = x.Rs;
			int cs = x.Cs;
			if (rs != z.Rs || cs != z.Cs)
				ErrIncompatible();
			else
			{
				MultiID.SetToSIS(x, z);
				for (int r = 0; r < rs; r++)
				{
					for (int c = 0; c < cs; c++)
					{
						MultiID.SetToSIS(x.E[r, c], z.E[r, c]);
					}
				}
			}
		}

		public static void NormM(MultiJ z) // normalize
		{
			Calc.NormNM(z);
			for (int r = 0; r < z.Rs; r++)
			{
				for (int c = 0; c < z.Cs; c++)
				{
					Calc.NormNM(z.E[r, c]);
				}
			}
		}

		public static void RAddM(MultiJ x, MultiJ z) // z = z + x
		// sets muErr true if overflow
		{
			AddM(z, x, z);
		}

		public static void RSubM(MultiJ x, MultiJ z) // z = z - x
		// sets muErr true if overflow
		{
			if (Object.ReferenceEquals(x, z))
				ClearM(ref z);
			else
			{
				ChSM(x);
				RAddM(x, z);
				ChSM(x);
			}		
		}

		public static void AddM(MultiJ x, MultiJ y, MultiJ z) // z = x + y
		// sets muErr true if overflow
		{
			if (x.Rs == 0 && y.Rs == 0)
				MultiID.AddSI(x, y, z);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				ErrIncompatible();
			else
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
						z.E[r, c] &= x.E[r, c] + y.E[r, c];
		}

		public static void SubM(MultiJ x, MultiJ y, MultiJ z) // z = x - y
		// sets muErr true if overflow
		{
			if (Object.ReferenceEquals(x, y))
				ClearM(ref z);
			else if (Object.ReferenceEquals(x, z))
				RSubM(y, z);
			else if (Object.ReferenceEquals(y, z))
			{
				RSubM(x, z);
				ChSM(z);
			}
			else
			{
				ChSM(y);
				AddM(x, y, z);
				ChSM(y);
			}
		}

		public static bool aDigit(string st, string maxD)	// true if ch is a digit <= maxD
		{
			char ch;
			ch = Char.ToUpper(st[0]);
			return (ch >= '0' && ch <= maxD[0] && (ch <= '9' || ch >= 'A'));
		}

		public static void SetTo1J(int d1, MultiJ z)	// z = d1
		// allows a literal d1, d1 an integer, d1 not changed
		{
			if (z.Rs > 0)
			{
				z.E = new MultiSI[0, 0];
				z.Rs = z.Cs = 0;
			}
			MultiID.SetTo1SI(d1, z);
		}

		public static void SetToExM(Int64 ext, MultiJ z)	// z = ext (Int64 = long)
		// allows a literal ext, ext not changed, max Int64 = 9223,37203,68547,75807 = 2^63-1
		{
			if (z.Rs > 0)
			{
				z.E = new MultiSI[0, 0];
				z.Rs = z.Cs = 0;
			}
			MultiID.SetToExSI(ext, z);
		}

		public static void SetToDbM(double db, MultiJ z) // z = db % baseM^4
		// allows a literal db, db not changed, muErr set true if overflow
		{
			if (z.Rs > 0)
			{
				z.E = new MultiSI[0, 0];
				z.Rs = z.Cs = 0;
			}
			MultiID.SetToDbSI(db, z);
		}

		public static void SetToDeM(decimal de, MultiJ z) // z = db
		// allows a literal de, de not changed, muErr set true if overflow
		{
			if (z.Rs > 0)
			{
				z.E = new MultiSI[0, 0];
				z.Rs = z.Cs = 0;
			}
			MultiID.SetToDeSI(de, z);
		}

		public static void Get1J(out int d1, MultiJ z) // d1 = z % baseM
		// z is not changed
		{
			d1 = 0;
			if (z.Rs != 0)
				ErrIncompatible();
			else
				MultiID.Get1SI(out d1, z);
		}

		public static void GetDeM(out decimal ext, MultiJ z)	// ext = z
		// z is not changed
		// max integer decimal is 2^96-1 = 7922,81625,14264,33759,35439,50335 29 digits
		{
			ext = 0;
			if (z.Rs != 0)
				ErrIncompatible();
			else
				MultiID.GetDeSI(out ext, z);
		}

		public static void GetDbM(out double db, MultiJ z)	// db = IntegerPart(z) % baseM^4
		// z is not changed, 2^53-1 = 9,00719,92547,40991 16 digits
		{
			db = 0;
			if (z.Rs != 0)
				ErrIncompatible();
			else
				MultiID.GetDbSI(out db, z);
		}

		public static void GetDbA(out double db, MultiJ z)	// db = IntegerPart(z) + FractionPart(z)
		// z is not changed
		{
			db = 0;
			if (z.Rs != 0)
				ErrIncompatible();
			else
				MultiID.GetDbSI(out db, z);
		}

		public static void GetDbInJ(out double db, double low, double high, MultiJ z)
		// db = z % baseM^4, in a range, + or -
		{
			db = 0;
			if (z.Rs != 0)
				ErrIncompatible();
			else
				MultiID.GetDbInSI(out db, low, high, z);
		}

		public static void RAdd1M(int d1, MultiJ z) // z = z + d1
		// allows literal d1, d1 is not changed, sets muErr true if overflow
		{
			if (z.Rs != 0)
				ErrIncompatible();
			else
				MultiID.RAdd1SI(d1, z);
		}

		public static void RSub1M(int d1, MultiJ z) // z = z - d1
		// allows literal d1, d1 is not changed, sets muErr true if overflow
		{
			if (z.Rs != 0)
				ErrIncompatible();
			else
				MultiID.RAdd1SI(-d1, z);
		}

		public static void Add1M(MultiJ x, int d1, MultiJ z) // z = x + d1
		// allows literal d1, d1 is not changed, sets muErr true if overflow
		{
			if (x.Rs != 0 || z.Rs != 0)
				ErrIncompatible();
			else
				MultiID.Add1SI(x, d1, z);
		}

		public static void WriteJ(MultiJ z, ref int tot)
		{
			WriteJ(MultiID.outS, null, z, ref tot, 0);
		}

		public static void WriteJ(RichTextBox outB, StreamWriter outF, MultiJ x, ref int tot, int width)
		{
			WriteJ(outB, outF, x, ref tot, width, false);
		}

		public static void WriteJ(RichTextBox outB, StreamWriter outF, MultiJ x, ref int tot, int width, bool WriteBase)
		// output x as a line of text, tot = characters on line so far, x is not modified
		// good for any base = baseO [2, 36], base 10 much faster
		{
			if (x.Rs == 0)
			{
				MultiID.WriteSI(outB, outF, x, ref tot, width, true);
			}
			else
			{
				bool wB = MultiID.writeBool;
				MultiID.writeBool = false;
				MultiID.WriteT(outB, outF, "{", ref tot, width);
				int rs = x.Rs;
				int cs = x.Cs;
				for (int r = 0; r < rs; r++)
				{
					for (int c = 0; c < cs; c++)
					{
						MultiID.MuInterrupt();		// allow for abort
						if (MultiID.muAbort) break;	// exit for
						MultiID.WriteSI(outB, outF, x.E[r, c], ref tot, width, true);
						if (MultiID.diagOn && Calc.listOn)
							MultiID.WriteT(" N=" + x.E[r, c].N + " M=" +
								x.E[r, c].M + " U=" + x.E[r, c].V.GetUpperBound(0), ref tot);
						if (r == rs-1 && c == cs-1)
							MultiID.WriteT(outB, outF, "}", ref tot, width);
						else
						{
							if (c == cs-1)
							{
								MultiID.WriteTLn(outB, outF, ";+", ref tot, width);
								MultiID.WriteTLn(outB, outF, "+", ref tot, width);
							}
							else
								MultiID.WriteTLn(outB, outF, ",+", ref tot, width);
						}
					}
					if (MultiID.muAbort) break;	// exit for
				}
				MultiID.writeBool = wB;
			}
		} // WriteJ

		public static void RMul1M(int d1, MultiJ z) // z = z * d1 (d1 <= maxDigit)
		// Multi-precision one super digit multiply d1 is not changed
		// muErr set true if z overflows
		{
			if (z.Rs == 0)
				MultiID.RMul1SI(d1, z);
			else
				for (int r = 0; r < z.Rs; r++)
					for (int c = 0; c < z.Cs; c++)
						MultiID.RMul1SI(d1, z.E[r, c]);
		}

		public static void Mul1M(MultiJ x, int d1, MultiJ z) // z = x * d1 (d1 <= maxDigit)
		// Multi-precision one super digit multiply x and d1 are not changed
		// muErr set true if z overflows
		{
			SetToM(x, z);
			RMul1M(d1, z);
		}

		public static MultiJ Diagonal(MultiJ x) // return diagonal matrix from vector x
		{
			if (x.Rs == 0)
				ErrNotMatrix();
			else if (x.Rs != 1 && x.Cs != 1)
				ErrIncompatible();
			else
			{
				int n = Math.Max(x.Rs, x.Cs);
				MultiJ t;
				Matrix(Calc.fMC, n, n, out t);
				for (int r = 0; r < n; r++)
					for (int c = 0; c < n; c++)
						t.E[r, c] &= MultiID.sZero;
				if (x.Rs == 1)
					for (int i = 0; i < n; i++)
						Calc.SetToNM(x.E[0, i], t.E[i, i]);
				else
					for (int i = 0; i < n; i++)
						Calc.SetToNM(x.E[i, 0], t.E[i, i]);
				return t;
			}
			return mZero;
		} // Diagonal

		public static void DiagonalM(MultiJ x, MultiJ z) // z = diagonal matrix from vector x
		{
			z &= Diagonal(x);
		} // DiagonalM

		public static void MulM(MultiJ x, MultiJ y, MultiJ z) // z = x * y
		// x or y or both may have the same address in memory as z
		// x, y are not changed unless they share memory with z
		// sets muErr true if overflow
		{
			if (x.Rs == 0 && y.Rs == 0)
				MultiID.MulSI(x, y, z);
			else if (x.Rs == 0)
			{
				MultiJ zM = new MultiJ(0, y.Rs, y.Cs);
				for (int r = 0; r < y.Rs; r++)
					for (int c = 0; c < y.Cs; c++)
						MultiID.SetToSI(x * y.E[r, c], zM.E[r, c]);
				SetToM(zM, z);
			}
			else if (y.Rs == 0)
			{
				MultiJ zM = new MultiJ(0, x.Rs, x.Cs);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
						MultiID.SetToSI(y * x.E[r, c], zM.E[r, c]);
				SetToM(zM, z);
			}
			else if (x.Cs != y.Rs)
				ErrIncompatible();
			else
			{
				MultiSI pF = new MultiSI();
				MultiJ zM = new MultiJ(0, x.Rs, y.Cs);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < y.Cs; c++)
					{
						pF &= 0;
						for (int i = 0; i < x.Cs; i++)
						{
							pF += x.E[r, i] * y.E[i, c];
						}
						zM.E[r, c] &= pF;
					}
				SetToM(zM, z);
			}
		} // MulM

		public static void RMulM(MultiJ x, MultiJ z) // z = z * x
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow
		{
			MulM(z, x, z);
		}

		public static void RSqM(MultiJ z) // z = z * z
		// sets muErr true if overflow
		{
			MulM(z, z, z);
		}

		public static void RDivMF(MultiSI x, MultiJ z) // z = z / x
		// sets muErr true if overflow
		{
			DivMF(z, x, z);
		}

		public static void RDiv1M(int d1, MultiJ z) // z = z / d1 (d1 <= maxDigit)
		// Multi-precision one super digit divide, d1 is not changed
		// allows literal d1, muErr set true if d1 = 0
		{
			MultiSI x = new MultiSI(2); // temp
			MultiID.SetTo1SI(d1, x);
			RDivMF(x, z);
		}

		public static void Div1M(MultiJ u, int d1, MultiJ z) // z = u / d1 (d1 <= maxDigit)
		// Multi-precision one super digit divide, d1 and u not changed
		// u and z may have same address in memory, then u is changed
		// allows literal d1, muErr set true if d1 = 0
		{
			SetToM(u, z);
			RDiv1M(d1, z);
		}

		public static void DivM(MultiJ u, MultiJ d, MultiJ z) // z = u / d
		// u or d or both may have the same address in memory as z
		// u, d are not changed unless they share memory with z
		// sets muErr true if d = 0
		{
			if (u.Rs == 0 && d.Rs == 0)
				MultiID.DivSI(u, d, z);
			else if (u.Rs == 0)
			{
				MultiJ zM = new MultiJ(Calc.fMC, d.Rs, d.Cs);
				for (int r = 0; r < d.Rs; r++)
					for (int c = 0; c < d.Cs; c++)
						MultiID.SetToSI(u / d.E[r, c], zM.E[r, c]);
				Calc.SetToNM(zM, z);
			}
			else if (d.Rs == 0)
			{
				MultiJ zM = new MultiJ(Calc.fMC, u.Rs, u.Cs);
				for (int r = 0; r < u.Rs; r++)
					for (int c = 0; c < u.Cs; c++)
						MultiID.SetToSI(u.E[r, c] / d, zM.E[r, c]);
				Calc.SetToNM(zM, z);
			}
			else if (u.Rs != d.Rs || u.Cs != d.Cs)
				ErrIncompatible();
			else
			{
				MultiJ zM = new MultiJ(Calc.fMC, u.Rs, u.Cs);
				for (int r = 0; r < u.Rs; r++)
					for (int c = 0; c < u.Cs; c++)
						MultiID.DivSI(u.E[r, c], d.E[r, c], zM.E[r, c]);
				Calc.SetToNM(zM, z);
			}
		} // DivM

		public static void DivMF(MultiJ u, MultiSI d, MultiJ z) // z = u / d
		// u or d or both may have the same address in memory as z
		// u, d are not changed unless they share memory with z
		// sets muErr true if d = 0
		{
			z &= (MultiSI)u / d;
			NormM(z);
		} // DivMF

		public static void RDivM(MultiJ d, MultiJ z) // z = z / d
		// d may have the same address in memory as z
		// d is not changed unless it shares memory with z
		// sets muErr true if d = 0
		{
			DivM(z, d, z);
		}

		public static bool ZTestM(MultiJ z) // true if z is Zero
		{
			if (z.Rs == 0)
				return (MultiID.ZTestSI(z));
			for (int r = 0; r < z.Rs; r++)
				for (int c = 0; c < z.Cs; c++)				
				{
					if (!MultiID.ZTestSI(z.E[r, c]))
						return false;
				}
			return true;
		}


		public static void ErrElementOfScalar()
		{
			MultiID.MuWriteErr("Cannot set an element of a scalar");
		}
	
		public static void ErrNotScalar()
		{
			MultiID.MuWriteErr("Cannot use a matrix as a scalar");
		}
	
		public static void ErrNotMatrix()
		{
			MultiID.MuWriteErr("Cannot use a scalar as a matrix");
		}
	
		public static void ErrOutside()
		{
			MultiID.MuWriteErr("Element outside of matrix");
		}
	
		public static void ErrIncompatible()
		{
			MultiID.MuWriteErr("Matrices are incompatible");
		}
	
		public static void ErrNotSquare()
		{
			MultiID.MuWriteErr("This is not a square matrix");
		}
	
		public static void ErrFMBNotScalar()
		{
			MultiID.MuWriteErr("FMB is not a scalar");
		}

		public static MultiJ mZero;		// mZero = 0 zero scalar
		public static MultiJ mOne;		// mOne  = 1 scalar
		public static MultiJ mTwo;		// mTwo  = 2 scalar
		public static MultiJ mTen;		// mTen  = 10 scalar
		public static MultiJ vZero;		// vZero = {0} zero vector
	}
}
